# -*- coding: utf-8 -*-

from pysnow.client import *
from pysnow.query import *
from pysnow.request import *
from pysnow.exceptions import *

__author__ = "Robert Wikman <rbw@vault13.org>"
__version__ = "0.4.10"



# -*- coding: utf-8 -*-

import inspect
from pysnow.exceptions import (QueryEmpty,
                               QueryExpressionError,
                               QueryMissingField,
                               QueryMultipleExpressions,
                               QueryTypeError)


class QueryBuilder(object):
    def __init__(self):
        """Query builder - used for building complex queries"""
        self._query = []
        self.current_field = None
        self.c_oper = None
        self.l_oper = None

    def AND(self):
        """And operator"""
        return self._add_logical_operator('^')

    def OR(self):
        """OR operator"""
        return self._add_logical_operator('^OR')

    def NQ(self):
        """NQ (new query) operator"""
        return self._add_logical_operator('^NQ')

    def field(self, field):
        """ Sets the field to operate on

        :param field: field (str) to operate on
        :return: self
        """
        self.current_field = field
        return self

    def starts_with(self, value):
        """Query records with the given field starting with the value specified"""
        return self._add_condition('STARTSWITH', value, types=[str])

    def ends_with(self, value):
        """Query records with the given field ending with the value specified"""
        return self._add_condition('ENDSWITH', value, types=[str])

    def contains(self, value):
        """Query records with the given field containing the value specified"""
        return self._add_condition('LIKE', value, types=[str])

    def not_contains(self, value):
        """Query records with the given field not containing the value specified"""
        return self._add_condition('NOTLIKE', value, types=[str])

    def is_empty(self):
        """Query records with the given field empty"""
        return self._add_condition('ISEMPTY', '', types=[str, int])

    def equals(self, value):
        """
        Query records with the given field equalling either:
        - the value passed (str)
        - any of the values passed (list)
        """

        if isinstance(value, str):
            return self._add_condition('=', value, types=[int, str])
        elif isinstance(value, list):
            return self._add_condition('IN', ",".join(value), types=[str])

        raise QueryTypeError('Expected value of type `str` or `list`, not %s' % type(value))

    def not_equals(self, value):
        """
        Query records with the given field not equalling:
        - the value specified
        - any of the values specified (list)
        """

        if isinstance(value, str):
            return self._add_condition('!=', value, types=[int, str])
        elif isinstance(value, list):
            return self._add_condition('NOT IN', ",".join(value), types=[str])

        raise QueryTypeError('Expected value of type `str` or `list`, not %s' % type(value))

    def greater_than(self, value):
        """Query records with the given field greater than the value specified"""
        if hasattr(value, 'strftime'):
            value = value.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(value, str):
            raise QueryTypeError('Expected value of type `int` or instance of `datetime`, not %s' % type(value))

        return self._add_condition('>', value, types=[int, str])

    def less_than(self, value):
        """Query records with the given field less than the value specified"""
        if hasattr(value, 'strftime'):
            value = value.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(value, str):
            raise QueryTypeError('Expected value of type `int` or instance of `datetime`, not %s' % type(value))

        return self._add_condition('<', value, types=[int, str])

    def between(self, start, end):
        """Query records in a start and end range

        :param start: `int` or `datetime` object
        :param end: `int` or `datetime` object
        :raise:
            :QueryTypeError: if start or end arguments is of an invalid type
        :return: self
        """
        if hasattr(start, 'strftime') and hasattr(end, 'strftime'):
            dt_between = (
              'javascript:gs.dateGenerate("%(start)s")'
              "@"
              'javascript:gs.dateGenerate("%(end)s")'
            ) % {
              'start': start.strftime('%Y-%m-%d %H:%M:%S'),
              'end': end.strftime('%Y-%m-%d %H:%M:%S')
            }
        elif isinstance(start, int) and isinstance(end, int):
            dt_between = '%d@%d' % (start, end)
        else:
            raise QueryTypeError("Expected `start` and `end` of type `int` "
                                 "or instance of `datetime`, not %s and %s" % (type(start), type(end)))

        return self._add_condition('BETWEEN', dt_between, types=[str])

    def _add_condition(self, operator, value, types):
        """ Appends condition to self._query after performing validation

        :param operator: operator (str)
        :param value: value / operand
        :param types: allowed types
        :raise:
            :QueryMissingField: if a field hasn't been set
            :QueryMultipleExpressions: if a condition already has been set
            :QueryTypeError: if the value is of an unexpected type
        :return: self
        """
        if not self.current_field:
            raise QueryMissingField("Expressions requires a field()")
        elif not type(value) in types:
            caller = inspect.currentframe().f_back.f_code.co_name
            raise QueryTypeError("Invalid type passed to %s() , expected: %s" % (caller, types))
        elif self.c_oper:
            raise QueryMultipleExpressions("Expected logical operator after expression")

        self.c_oper = inspect.currentframe().f_back.f_code.co_name
        self._query.append("%(current_field)s%(operator)s%(value)s" % {
                               'current_field': self.current_field,
                               'operator': operator,
                               'value': value})
        return self

    def _add_logical_operator(self, operator):
        """ Adds a logical operator between conditions in query

        :param operator: logical operator (str)
        :raise:
            :QueryExpressionError: if a expression hasn't been set
        :return: self
        """
        if not self.c_oper:
            raise QueryExpressionError("Logical operators must be preceded by an expression")

        self.current_field = None
        self.c_oper = None

        self.l_oper = inspect.currentframe().f_back.f_code.co_name
        self._query.append(operator)
        return self

    def __str__(self):
        """ String representation of the query object
        :raise:
            :QueryEmpty: if there's no conditions defined
            :QueryMissingField: if field() hasn't been set
            :QueryExpressionError: if a expression hasn't been set
        :return: Query string
        """
        if len(self._query) == 0:
            raise QueryEmpty("At least one condition is required")
        elif self.current_field is None:
            raise QueryMissingField("Logical operator expects a field()")
        elif self.c_oper is None:
            raise QueryExpressionError("field() expects an expression")

        return str().join(self._query)


# For backwards compatibility
class Query(QueryBuilder):
    pass



# -*- coding: utf-8 -*-

import requests
import warnings
from pysnow import request
from pysnow.exceptions import InvalidUsage

warnings.simplefilter("always", DeprecationWarning)


class Client(object):
    def __init__(self,
                 instance=None,
                 host=None,
                 user=None,
                 password=None,
                 raise_on_empty=True,
                 default_payload=None,
                 request_params=None,
                 use_ssl=True,
                 session=None):
        """Sets configuration and creates a session object used in `Request` later on

        You must either provide a username and password or a requests session.
        If you provide a requests session it must handle the authentication.
        For example, providing a session can be used to do OAuth authentication.

        :param instance: instance name, used to construct host
        :param host: host can be passed as an alternative to instance
        :param user: username
        :param password: password
        :param raise_on_empty: whether or not to raise an exception on 404 (no matching records)
        :param default_payload: deprecated, use request_params
        :param request_params: request params to send with requests
        :param use_ssl: Enable or disable SSL
        :param session: a requests session object
        """

        # We allow either host or instance, not both
        if (host and instance) is not None:
            raise InvalidUsage("Got both 'instance' and 'host'. Panic.")

        if ((not (user and password)) and not session) or ((user or password) and session):
            raise InvalidUsage("You must either provide username and password or a session")

        # Check if host or instance was passed
        if instance:
            self.host = "%s.service-now.com" % instance
        elif host:
            self.host = host
        else:
            raise InvalidUsage("You must pass 'instance' or 'host' to create a client")

        # Check if SSL was requested
        if use_ssl is True:
            self.base_url = "https://%s" % self.host
        elif use_ssl is False:
            self.base_url = "http://%s" % self.host
        else:
            raise InvalidUsage("use_ssl: boolean value expected")

        # Connection properties
        self.instance = instance
        self._user = user
        self._password = password
        self.raise_on_empty = raise_on_empty
        self.request_params = self.default_payload = {}

        # default_payload is deprecated, let user know
        if default_payload is not None:
            warnings.warn("default_payload is deprecated, please use request_params instead", DeprecationWarning)
            self.request_params = self.default_payload = default_payload

        if request_params is not None:
            self.request_params = request_params

        # Sets request parameters for requests
        if not isinstance(self.request_params, dict):
            raise InvalidUsage("Payload must be of type dict")

        # Create new session object
        self.session = self._get_session(session)

    def _get_session(self, session):
        """Creates and returns a new session object with the credentials passed to the constructor

        :return: session object
        """
        if session:
            s = session
        else:
            s = requests.Session()
            s.auth = requests.auth.HTTPBasicAuth(self._user, self._password)

        s.headers.update({'content-type': 'application/json', 'accept': 'application/json'})
        return s

    def _request(self, method, table, **kwargs):
        """Creates and returns a new `Request` object, takes some basic settings from the `Client` object and
        passes along to the `Request` constructor

        :param method: HTTP method
        :param table: Table to operate on
        :param kwargs: Keyword arguments passed along to `Request`
        :return: `Request` object
        """
        return request.Request(method,
                               table,
                               request_params=self.request_params,
                               raise_on_empty=self.raise_on_empty,
                               session=self.session,
                               instance=self.instance,
                               base_url=self.base_url,
                               **kwargs)

    def query(self, table, **kwargs):
        """Query (GET) request wrapper.

        :param table: table to perform query on
        :param kwargs: Keyword arguments passed along to `Request`
        :return: `Request` object
        """
        return self._request('GET', table, **kwargs)

    def insert(self, table, payload, **kwargs):
        """Insert (POST) request wrapper

        :param table: table to insert on
        :param payload: update payload (dict)
        :param kwargs: Keyword arguments passed along to `Request`
        :return: New record content
        """
        r = self._request('POST', table, **kwargs)
        return r.insert(payload)




# -*- coding: utf-8 -*-

import io
import ast

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version():
    """ Parses pysnow.py and fetches the version attribute from the syntax tree
    :return: pysnow version
    """
    with io.open('pysnow/__init__.py') as input_file:
        for line in input_file:
            if line.startswith('__version__'):
                return ast.parse(line).body[0].value.s

with io.open('README.rst') as readme:
    setup(
        name='pysnow',
        packages=['pysnow'],
        version=get_version(),
        description='Python library for the ServiceNow REST API',
        long_description=readme.read(),
        install_requires=['requests'],
        author='Robert Wikman',
        author_email='rbw@vault13.org',
        maintainer='Robert Wikman',
        maintainer_email='rbw@vault13.org',
        url='https://github.com/rbw0/pysnow',
        download_url='https://github.com/rbw0/pysnow/tarball/%s' % get_version(),
        keywords=['servicenow', 'rest', 'api', 'http'],
        platforms='any',
        classifiers=[
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Software Development :: Libraries :: Python Modules'
        ],
        license='MIT',
    )